// --------------------------------------------------------------------
// Copyright (c) 2005 by Terasic Technologies Inc. 
// --------------------------------------------------------------------
//
// Permission:
//
//   Terasic grants permission to use and modify this code for use
//   in synthesis for all Terasic Development Boards and Altera Development 
//   Kits made by Terasic.  Other use of this code, including the selling 
//   ,duplication, or modification of any portion is strictly prohibited.
//
// Disclaimer:
//
//   This VHDL/Verilog or C/C++ source code is intended as a design reference
//   which illustrates how these types of functions can be implemented.
//   It is the user's responsibility to verify their design for
//   consistency and functionality through the use of formal
//   verification methods.  Terasic provides no warranty regarding the use 
//   or functionality of this code.
//
// --------------------------------------------------------------------
//           
//                     Terasic Technologies Inc
//                     356 Fu-Shin E. Rd Sec. 1. JhuBei City,
//                     HsinChu County, Taiwan
//                     302
//
//                     web: http://www.terasic.com/
//                     email: support@terasic.com
//
// --------------------------------------------------------------------
// Major Functions:	DE2 TV Box
// --------------------------------------------------------------------
//
// Revision History :
// --------------------------------------------------------------------
//   Ver  :| Author            :| Mod. Date :| Changes Made:
//   V1.0 :| Joe Yang	       :| 05/07/05  :| Initial Revision
//   V1.1 :| Johnny Chen       :| 05/09/05  :| Changed YCbCr2RGB Block,
//											   RGB output 8 Bits => 10 Bits
//   V1.2 :| Johnny Chen	   :| 05/10/05  :| H_SYNC & V_SYNC Timing fixed.
//   V1.3 :| Johnny Chen       :| 05/11/16  :| Added FLASH Address FL_ADDR[21:20]
//   V1.4 :| Joe Yang	       :| 06/07/20  :| Modify Output Color
//	 V2.0 :| Johnny Chen	   :| 06/11/20	:| New Version for DE2 v2.X PCB.
// --------------------------------------------------------------------

module DE2_TV
	(
		////////////////////	Clock Input	 	////////////////////	 
		CLOCK_27,						//	27 MHz
		CLOCK_50,						//	50 MHz
		EXT_CLOCK,						//	External Clock
		////////////////////	Push Button		////////////////////
		KEY,							//	Pushbutton[3:0]
		////////////////////	DPDT Switch		////////////////////
		SW,								//	Toggle Switch[17:0]
		////////////////////	7-SEG Dispaly	////////////////////
		HEX0,							//	Seven Segment Digit 0
		HEX1,							//	Seven Segment Digit 1
		HEX2,							//	Seven Segment Digit 2
		HEX3,							//	Seven Segment Digit 3
		HEX4,							//	Seven Segment Digit 4
		HEX5,							//	Seven Segment Digit 5
		HEX6,							//	Seven Segment Digit 6
		HEX7,							//	Seven Segment Digit 7
		////////////////////////	LED		////////////////////////
		LEDG,							//	LED Green[8:0]
		LEDR,							//	LED Red[17:0]
		////////////////////////	UART	////////////////////////
		UART_TXD,						//	UART Transmitter
		UART_RXD,						//	UART Receiver
		////////////////////////	IRDA	////////////////////////
		IRDA_TXD,						//	IRDA Transmitter
		IRDA_RXD,						//	IRDA Receiver
		/////////////////////	SDRAM Interface		////////////////
		DRAM_DQ,						//	SDRAM Data bus 16 Bits
		DRAM_ADDR,						//	SDRAM Address bus 12 Bits
		DRAM_LDQM,						//	SDRAM Low-byte Data Mask 
		DRAM_UDQM,						//	SDRAM High-byte Data Mask
		DRAM_WE_N,						//	SDRAM Write Enable
		DRAM_CAS_N,						//	SDRAM Column Address Strobe
		DRAM_RAS_N,						//	SDRAM Row Address Strobe
		DRAM_CS_N,						//	SDRAM Chip Select
		DRAM_BA_0,						//	SDRAM Bank Address 0
		DRAM_BA_1,						//	SDRAM Bank Address 0
		DRAM_CLK,						//	SDRAM Clock
		DRAM_CKE,						//	SDRAM Clock Enable
		////////////////////	Flash Interface		////////////////
		FL_DQ,							//	FLASH Data bus 8 Bits
		FL_ADDR,						//	FLASH Address bus 22 Bits
		FL_WE_N,						//	FLASH Write Enable
		FL_RST_N,						//	FLASH Reset
		FL_OE_N,						//	FLASH Output Enable
		FL_CE_N,						//	FLASH Chip Enable
		////////////////////	SRAM Interface		////////////////
		SRAM_DQ,						//	SRAM Data bus 16 Bits
		SRAM_ADDR,						//	SRAM Address bus 18 Bits
		SRAM_UB_N,						//	SRAM High-byte Data Mask 
		SRAM_LB_N,						//	SRAM Low-byte Data Mask 
		SRAM_WE_N,						//	SRAM Write Enable
		SRAM_CE_N,						//	SRAM Chip Enable
		SRAM_OE_N,						//	SRAM Output Enable
		////////////////////	ISP1362 Interface	////////////////
		OTG_DATA,						//	ISP1362 Data bus 16 Bits
		OTG_ADDR,						//	ISP1362 Address 2 Bits
		OTG_CS_N,						//	ISP1362 Chip Select
		OTG_RD_N,						//	ISP1362 Write
		OTG_WR_N,						//	ISP1362 Read
		OTG_RST_N,						//	ISP1362 Reset
		OTG_FSPEED,						//	USB Full Speed,	0 = Enable, Z = Disable
		OTG_LSPEED,						//	USB Low Speed, 	0 = Enable, Z = Disable
		OTG_INT0,						//	ISP1362 Interrupt 0
		OTG_INT1,						//	ISP1362 Interrupt 1
		OTG_DREQ0,						//	ISP1362 DMA Request 0
		OTG_DREQ1,						//	ISP1362 DMA Request 1
		OTG_DACK0_N,					//	ISP1362 DMA Acknowledge 0
		OTG_DACK1_N,					//	ISP1362 DMA Acknowledge 1
		////////////////////	LCD Module 16X2		////////////////
		LCD_ON,							//	LCD Power ON/OFF
		LCD_BLON,						//	LCD Back Light ON/OFF
		LCD_RW,							//	LCD Read/Write Select, 0 = Write, 1 = Read
		LCD_EN,							//	LCD Enable
		LCD_RS,							//	LCD Command/Data Select, 0 = Command, 1 = Data
		LCD_DATA,						//	LCD Data bus 8 bits
		////////////////////	SD_Card Interface	////////////////
		SD_DAT,							//	SD Card Data
		SD_DAT3,						//	SD Card Data 3
		SD_CMD,							//	SD Card Command Signal
		SD_CLK,							//	SD Card Clock
		////////////////////	USB JTAG link	////////////////////
		TDI,  							// CPLD -> FPGA (data in)
		TCK,  							// CPLD -> FPGA (clk)
		TCS,  							// CPLD -> FPGA (CS)
	    TDO,  							// FPGA -> CPLD (data out)
		////////////////////	I2C		////////////////////////////
		I2C_SDAT,						//	I2C Data
		I2C_SCLK,						//	I2C Clock
		////////////////////	PS2		////////////////////////////
		PS2_DAT,						//	PS2 Data
		PS2_CLK,						//	PS2 Clock
		////////////////////	VGA		////////////////////////////
		VGA_CLK,   						//	VGA Clock
		VGA_HS,							//	VGA H_SYNC
		VGA_VS,							//	VGA V_SYNC
		VGA_BLANK,						//	VGA BLANK
		VGA_SYNC,						//	VGA SYNC
		VGA_R,   						//	VGA Red[9:0]
		VGA_G,	 						//	VGA Green[9:0]
		VGA_B,  						//	VGA Blue[9:0]
		////////////	Ethernet Interface	////////////////////////
		ENET_DATA,						//	DM9000A DATA bus 16Bits
		ENET_CMD,						//	DM9000A Command/Data Select, 0 = Command, 1 = Data
		ENET_CS_N,						//	DM9000A Chip Select
		ENET_WR_N,						//	DM9000A Write
		ENET_RD_N,						//	DM9000A Read
		ENET_RST_N,						//	DM9000A Reset
		ENET_INT,						//	DM9000A Interrupt
		ENET_CLK,						//	DM9000A Clock 25 MHz
		////////////////	Audio CODEC		////////////////////////
		AUD_ADCLRCK,					//	Audio CODEC ADC LR Clock
		AUD_ADCDAT,						//	Audio CODEC ADC Data
		AUD_DACLRCK,					//	Audio CODEC DAC LR Clock
		AUD_DACDAT,						//	Audio CODEC DAC Data
		AUD_BCLK,						//	Audio CODEC Bit-Stream Clock
		AUD_XCK,						//	Audio CODEC Chip Clock
		////////////////	TV Decoder		////////////////////////
		TD_DATA,    					//	TV Decoder Data bus 8 bits
		TD_HS,							//	TV Decoder H_SYNC
		TD_VS,							//	TV Decoder V_SYNC
		TD_RESET,						//	TV Decoder Reset
		TD_CLK,							//	TV Decoder Line Locked Clock
		////////////////////	GPIO	////////////////////////////
		GPIO_0,							//	GPIO Connection 0
		GPIO_1							//	GPIO Connection 1
	);

////////////////////////	Clock Input	 	////////////////////////
input			CLOCK_27;				//	27 MHz
input			CLOCK_50;				//	50 MHz
input			EXT_CLOCK;				//	External Clock
////////////////////////	Push Button		////////////////////////
input	[3:0]	KEY;					//	Pushbutton[3:0]
////////////////////////	DPDT Switch		////////////////////////
input	[17:0]	SW;						//	Toggle Switch[17:0]
////////////////////////	7-SEG Dispaly	////////////////////////
output	[6:0]	HEX0;					//	Seven Segment Digit 0
output	[6:0]	HEX1;					//	Seven Segment Digit 1
output	[6:0]	HEX2;					//	Seven Segment Digit 2
output	[6:0]	HEX3;					//	Seven Segment Digit 3
output	[6:0]	HEX4;					//	Seven Segment Digit 4
output	[6:0]	HEX5;					//	Seven Segment Digit 5
output	[6:0]	HEX6;					//	Seven Segment Digit 6
output	[6:0]	HEX7;					//	Seven Segment Digit 7
////////////////////////////	LED		////////////////////////////
output	[8:0]	LEDG;					//	LED Green[8:0]
output	[17:0]	LEDR;					//	LED Red[17:0]
////////////////////////////	UART	////////////////////////////
output			UART_TXD;				//	UART Transmitter
input			UART_RXD;				//	UART Receiver
////////////////////////////	IRDA	////////////////////////////
output			IRDA_TXD;				//	IRDA Transmitter
input			IRDA_RXD;				//	IRDA Receiver
///////////////////////		SDRAM Interface	////////////////////////
inout	[15:0]	DRAM_DQ;				//	SDRAM Data bus 16 Bits
output	[11:0]	DRAM_ADDR;				//	SDRAM Address bus 12 Bits
output			DRAM_LDQM;				//	SDRAM Low-byte Data Mask 
output			DRAM_UDQM;				//	SDRAM High-byte Data Mask
output			DRAM_WE_N;				//	SDRAM Write Enable
output			DRAM_CAS_N;				//	SDRAM Column Address Strobe
output			DRAM_RAS_N;				//	SDRAM Row Address Strobe
output			DRAM_CS_N;				//	SDRAM Chip Select
output			DRAM_BA_0;				//	SDRAM Bank Address 0
output			DRAM_BA_1;				//	SDRAM Bank Address 0
output			DRAM_CLK;				//	SDRAM Clock
output			DRAM_CKE;				//	SDRAM Clock Enable
////////////////////////	Flash Interface	////////////////////////
inout	[7:0]	FL_DQ;					//	FLASH Data bus 8 Bits
output	[21:0]	FL_ADDR;				//	FLASH Address bus 22 Bits
output			FL_WE_N;				//	FLASH Write Enable
output			FL_RST_N;				//	FLASH Reset
output			FL_OE_N;				//	FLASH Output Enable
output			FL_CE_N;				//	FLASH Chip Enable
////////////////////////	SRAM Interface	////////////////////////
inout	[15:0]	SRAM_DQ;				//	SRAM Data bus 16 Bits
output	[17:0]	SRAM_ADDR;				//	SRAM Address bus 18 Bits
output			SRAM_UB_N;				//	SRAM High-byte Data Mask 
output			SRAM_LB_N;				//	SRAM Low-byte Data Mask 
output			SRAM_WE_N;				//	SRAM Write Enable
output			SRAM_CE_N;				//	SRAM Chip Enable
output			SRAM_OE_N;				//	SRAM Output Enable
////////////////////	ISP1362 Interface	////////////////////////
inout	[15:0]	OTG_DATA;				//	ISP1362 Data bus 16 Bits
output	[1:0]	OTG_ADDR;				//	ISP1362 Address 2 Bits
output			OTG_CS_N;				//	ISP1362 Chip Select
output			OTG_RD_N;				//	ISP1362 Write
output			OTG_WR_N;				//	ISP1362 Read
output			OTG_RST_N;				//	ISP1362 Reset
output			OTG_FSPEED;				//	USB Full Speed,	0 = Enable, Z = Disable
output			OTG_LSPEED;				//	USB Low Speed, 	0 = Enable, Z = Disable
input			OTG_INT0;				//	ISP1362 Interrupt 0
input			OTG_INT1;				//	ISP1362 Interrupt 1
input			OTG_DREQ0;				//	ISP1362 DMA Request 0
input			OTG_DREQ1;				//	ISP1362 DMA Request 1
output			OTG_DACK0_N;			//	ISP1362 DMA Acknowledge 0
output			OTG_DACK1_N;			//	ISP1362 DMA Acknowledge 1
////////////////////	LCD Module 16X2	////////////////////////////
inout	[7:0]	LCD_DATA;				//	LCD Data bus 8 bits
output			LCD_ON;					//	LCD Power ON/OFF
output			LCD_BLON;				//	LCD Back Light ON/OFF
output			LCD_RW;					//	LCD Read/Write Select, 0 = Write, 1 = Read
output			LCD_EN;					//	LCD Enable
output			LCD_RS;					//	LCD Command/Data Select, 0 = Command, 1 = Data
////////////////////	SD Card Interface	////////////////////////
inout			SD_DAT;					//	SD Card Data
inout			SD_DAT3;				//	SD Card Data 3
inout			SD_CMD;					//	SD Card Command Signal
output			SD_CLK;					//	SD Card Clock
////////////////////////	I2C		////////////////////////////////
inout			I2C_SDAT;				//	I2C Data
output			I2C_SCLK;				//	I2C Clock
////////////////////////	PS2		////////////////////////////////
input		 	PS2_DAT;				//	PS2 Data
input			PS2_CLK;				//	PS2 Clock
////////////////////	USB JTAG link	////////////////////////////
input  			TDI;					// CPLD -> FPGA (data in)
input  			TCK;					// CPLD -> FPGA (clk)
input  			TCS;					// CPLD -> FPGA (CS)
output 			TDO;					// FPGA -> CPLD (data out)
////////////////////////	VGA			////////////////////////////
output			VGA_CLK;   				//	VGA Clock
output			VGA_HS;					//	VGA H_SYNC
output			VGA_VS;					//	VGA V_SYNC
output			VGA_BLANK;				//	VGA BLANK
output			VGA_SYNC;				//	VGA SYNC
output	[9:0]	VGA_R;   				//	VGA Red[9:0]
output	[9:0]	VGA_G;	 				//	VGA Green[9:0]
output	[9:0]	VGA_B;   				//	VGA Blue[9:0]
////////////////	Ethernet Interface	////////////////////////////
inout	[15:0]	ENET_DATA;				//	DM9000A DATA bus 16Bits
output			ENET_CMD;				//	DM9000A Command/Data Select, 0 = Command, 1 = Data
output			ENET_CS_N;				//	DM9000A Chip Select
output			ENET_WR_N;				//	DM9000A Write
output			ENET_RD_N;				//	DM9000A Read
output			ENET_RST_N;				//	DM9000A Reset
input			ENET_INT;				//	DM9000A Interrupt
output			ENET_CLK;				//	DM9000A Clock 25 MHz
////////////////////	Audio CODEC		////////////////////////////
inout			AUD_ADCLRCK;			//	Audio CODEC ADC LR Clock
input			AUD_ADCDAT;				//	Audio CODEC ADC Data
inout			AUD_DACLRCK;			//	Audio CODEC DAC LR Clock
output			AUD_DACDAT;				//	Audio CODEC DAC Data
inout			AUD_BCLK;				//	Audio CODEC Bit-Stream Clock
output			AUD_XCK;				//	Audio CODEC Chip Clock
////////////////////	TV Devoder		////////////////////////////
input	[7:0]	TD_DATA;    			//	TV Decoder Data bus 8 bits
input			TD_HS;					//	TV Decoder H_SYNC
input			TD_VS;					//	TV Decoder V_SYNC
output			TD_RESET;				//	TV Decoder Reset
input			TD_CLK;					//	TV Decoder Line Locked Clock
////////////////////////	GPIO	////////////////////////////////
inout	[35:0]	GPIO_0;					//	GPIO Connection 0
inout	[35:0]	GPIO_1;					//	GPIO Connection 1

assign	LCD_ON		=	1'b1;
assign	LCD_BLON	=	1'b1;

//	All inout port turn to tri-state
assign	FL_DQ		=	8'hzz;
assign	SRAM_DQ		=	16'hzzzz;
assign	OTG_DATA	=	16'hzzzz;
assign	LCD_DATA	=	8'hzz;
assign	SD_DAT		=	1'bz;
assign	I2C_SDAT	=	1'bz;
assign	ENET_DATA	=	16'hzzzz;
assign	AUD_ADCLRCK	=	1'bz;
assign	AUD_DACLRCK	=	1'bz;
assign	AUD_BCLK	=	1'bz;

//	Enable TV Decoder
assign	TD_RESET	=	KEY[0];

wire		SYS_CLK;

assign	LEDG	=	VGA_Y;
assign	LEDR		=	VGA_X;

// Communication with central board
wire R_in;
wire G_in;
wire B_in;
reg R_in_d;
reg G_in_d;
reg B_in_d;

assign GPIO_1[9:0] = VGA_X[9:0];
assign GPIO_1[18:10] = VGA_Y[8:0];
assign GPIO_1[25:19] = detect1_x_mirrored;
assign GPIO_1[31:26] = detect1_y[5:0];
assign GPIO_1[32] = VGA_CLK;

assign B_in = GPIO_1[33];
assign G_in = GPIO_1[34];
assign R_in = GPIO_1[35];

always @ (posedge VGA_CLK) begin
   R_in_d <= R_in;
   B_in_d <= B_in;
   G_in_d <= G_in;
end

//	7 segment LUT
SEG7_LUT_8 			u0	(	.oSEG0(HEX0),
							.oSEG1(HEX1),
							.oSEG2(HEX2),
							.oSEG3(HEX3),
							.oSEG4(HEX4),
							.oSEG5(HEX5),
							.oSEG6(HEX6),
							.oSEG7(HEX7),
							.iDIG(SW) );

//	Audio CODEC and video decoder setting
I2C_AV_Config 		u1	(	//	Host Side
							.iCLK(CLOCK_50),
							.iRST_N(KEY[0]),
							//	I2C Side
							.I2C_SCLK(I2C_SCLK),
							.I2C_SDAT(I2C_SDAT)	);

//	TV Decoder Stable Check
TD_Detect			u2	(	.oTD_Stable(TD_Stable),
							.iTD_VS(TD_VS),
							.iTD_HS(TD_HS),
							.iRST_N(KEY[0])	);

//	Reset Delay Timer
Reset_Delay			u3	(	.iCLK(CLOCK_50),
							.iRST(TD_Stable),
							.oRST_0(DLY0),
							.oRST_1(DLY1),
							.oRST_2(DLY2));

//	ITU-R 656 to YUV 4:2:2
ITU_656_Decoder		u4	(	//	TV Decoder Input
							.iTD_DATA(TD_DATA),
							//	Position Output
							.oTV_X(TV_X),
							//	YUV 4:2:2 Output
							.oYCbCr(YCbCr),
							.oDVAL(TV_DVAL),
							//	Control Signals
							.iSwap_CbCr(Quotient[0]),
							.iSkip(Remain==4'h0),
							.iRST_N(DLY1),
							.iCLK_27(TD_CLK)	);

//	For Down Sample 720 to 640
DIV 				u5	(	.aclr(!DLY0),	
							.clock(TD_CLK),
							.denom(4'h9),
							.numer(TV_X),
							.quotient(Quotient),
							.remain(Remain));

//	SDRAM frame buffer
Sdram_Control_4Port	u6	(	//	HOST Side
						    .REF_CLK(CLOCK_27),
							.SYS_CLK(SYS_CLK),
						    .RESET_N(KEY[0]),
							//	FIFO Write Side 1
						    .WR1_DATA(YCbCr),
							.WR1(TV_DVAL),
							.WR1_FULL(WR1_FULL),
							.WR1_ADDR(0),
							.WR1_MAX_ADDR(640*507),		//	525-18
							.WR1_LENGTH(9'h80),
							.WR1_LOAD(!DLY0),
							.WR1_CLK(TD_CLK),
							//	FIFO Read Side 1
						    .RD1_DATA(m1YCbCr),
				        	.RD1(m1VGA_Read),
				        	.RD1_ADDR(640*13),			//	Read odd field and bypess blanking
							.RD1_MAX_ADDR(640*253),
							.RD1_LENGTH(9'h80),
				        	.RD1_LOAD(!DLY0),
							.RD1_CLK(VGA_CLK),
							//	FIFO Read Side 2
						    .RD2_DATA(m2YCbCr),
				        	.RD2(m2VGA_Read),
				        	.RD2_ADDR(640*267),			//	Read even field and bypess blanking
							.RD2_MAX_ADDR(640*507),
							.RD2_LENGTH(9'h80),
				        	.RD2_LOAD(!DLY0),
							.RD2_CLK(VGA_CLK),
							//	SDRAM Side
						    .SA(DRAM_ADDR),
						    .BA({DRAM_BA_1,DRAM_BA_0}),
						    .CS_N(DRAM_CS_N),
						    .CKE(DRAM_CKE),
						    .RAS_N(DRAM_RAS_N),
				            .CAS_N(DRAM_CAS_N),
				            .WE_N(DRAM_WE_N),
						    .DQ(DRAM_DQ),
				            .DQM({DRAM_UDQM,DRAM_LDQM}),
							.SDR_CLK(DRAM_CLK)	);

//	YUV 4:2:2 to YUV 4:4:4
YUV422_to_444		u7	(	//	YUV 4:2:2 Input
							.iYCbCr(mYCbCr),
							//	YUV	4:4:4 Output
							.oY(mY),
							.oCb(mCb),
							.oCr(mCr),
							//	Control Signals
							.iX(VGA_X),
							.iCLK(VGA_CLK),
							.iRST_N(DLY0));

//	YCbCr 8-bit to RGB-10 bit 
YCbCr2RGB 			u8	(	//	Output Side
							.Red(mRed),
							.Green(mGreen),
							.Blue(mBlue),
							.oDVAL(mDVAL),
							//	Input Side
							.iY(mY),
							.iCb(mCb),
							.iCr(mCr),
							.iDVAL(VGA_Read),
							//	Control Signal
							.iRESET(!DLY0),
							.iCLK(VGA_CLK));

//	VGA Controller
VGA_Controller u9(
    // Control Signal
    .iCLK (SYS_CLK),         // 50.4MHz Clock (PLL from 27MHz clock).
    .iRESET (~DLY2),       // System reset signal.
    // Host Side
    .iRed(SW[0] ? srVGA_R : SW[1] ? red_detbox1 : mRed),
    .iGreen(SW[0] ? srVGA_G : SW[1] ? green_detbox1 : mGreen),
    .iBlue(SW[0] ? srVGA_B : SW[1] ? blue_detbox1 : mBlue),
    .oAddress (Background_Addr),     // Address of pixel request
    .oCoord_X (VGA_X),     // X of pixel request (0 to 639, left is 0).
    .oCoord_Y (VGA_Y),     // Y of pixel request (0 to 479, top is 0)
    .oValidReq (VGA_Read),    // Pixel request is valid (rest of circuit
    // VGA Side (connect directly to VGA output signals)
    .oVGA_R(VGA_R),
    .oVGA_G(VGA_G),
    .oVGA_B(VGA_B),
    .oVGA_H_SYNC(VGA_HS),
    .oVGA_V_SYNC(VGA_VS),
    .oVGA_SYNC(VGA_SYNC),
    .oVGA_BLANK(VGA_BLANK)
);
reg VGA_CLK;
always @ (posedge SYS_CLK) VGA_CLK <= ~VGA_CLK;

//	For ITU-R 656 Decoder
wire	[15:0]	YCbCr;
wire	[9:0]	TV_X;
wire			TV_DVAL;

//	For VGA Controller
wire	[9:0]	mRed;
wire	[9:0]	mGreen;
wire	[9:0]	mBlue;
wire	[9:0]	VGA_X;
wire	[9:0]	VGA_Y;
wire			VGA_Read;	//	VGA data request
wire			m1VGA_Read;	//	Read odd field
wire			m2VGA_Read;	//	Read even field

//	For YUV 4:2:2 to YUV 4:4:4
wire	[7:0]	mY;
wire	[7:0]	mCb;
wire	[7:0]	mCr;

//	For field select
wire	[15:0]	mYCbCr;
wire	[15:0]	mYCbCr_d;
wire	[15:0]	m1YCbCr;
wire	[15:0]	m2YCbCr;
wire	[15:0]	m3YCbCr;

//	For Delay Timer
wire			TD_Stable;
wire			DLY0;
wire			DLY1;
wire			DLY2;

//	For Down Sample
wire	[3:0]	Remain;
wire	[9:0]	Quotient;

assign	m1VGA_Read	=	VGA_Y[0]		?	1'b0		:	VGA_Read	;
assign	m2VGA_Read	=	VGA_Y[0]		?	VGA_Read	:	1'b0		;
assign	mYCbCr_d	=	!VGA_Y[0]		?	m1YCbCr		:
											m2YCbCr		;
assign	mYCbCr		=	m5YCbCr;

wire			mDVAL;

//	Line buffer, delay one line
Line_Buffer u10	(	.clken(VGA_Read),
					.clock(VGA_CLK),
					.shiftin(mYCbCr_d),
					.shiftout(m3YCbCr));

Line_Buffer u11	(	.clken(VGA_Read),
					.clock(VGA_CLK),
					.shiftin(m3YCbCr),
					.shiftout(m4YCbCr));

wire	[15:0]	m4YCbCr;
wire	[15:0]	m5YCbCr;
wire	[8:0]	Tmp1,Tmp2;
wire	[7:0]	Tmp3,Tmp4;

assign	Tmp1	=	m4YCbCr[7:0]+mYCbCr_d[7:0];
assign	Tmp2	=	m4YCbCr[15:8]+mYCbCr_d[15:8];
assign	Tmp3	=	Tmp1[8:2]+m3YCbCr[7:1];
assign	Tmp4	=	Tmp2[8:2]+m3YCbCr[15:9];
assign	m5YCbCr	=	{Tmp4,Tmp3};

// For background
wire [19:0] Background_Addr;
reg [19:0] Background_Addr_d;
//assign Background_Addr = VGA_X + 640 * VGA_Y;
wire [9:0] Background_Red, Background_Green, Background_Blue;

reg BACKGROUND_CLK;
always @ (*) begin
   #3 BACKGROUND_CLK = ~SYS_CLK;
end
Video_Background bger	(
					.CLK(BACKGROUND_CLK),
					.address(Background_Addr),
					.red(Background_Red),
					.green(Background_Green),
					.blue(Background_Blue));

/*******************************************
 * PIP Code - Piggyback off VGA coords     *
 *******************************************/

reg [9:0] siX_Cont, siY_Cont;
assign SRAM_UB_N = 1'b0;
assign SRAM_LB_N = 1'b0;
assign SRAM_CE_N = 1'b0;
assign SRAM_OE_N = 1'b0;

wire srd_req, swr_req;
wire [9:0]  sram_raddr_x, sram_raddr_y;
parameter [9:0] x_max = 10'd159,
                y_max = 10'd119;
// assign swr_req = (mCCD_DVAL_d) && (siX_Cont <= x_max) && (siY_Cont <= y_max); // Write top left corner
// Get a scaled down picture, 1/8 size of original, and 4:3 aspect ratio
// from 640 x 512 to 160 x 120
assign swr_req = (VGA_X[1:0] == 0) && (VGA_Y[1:0] == 0) && (VGA_Y <= 10'd479);

assign SRAM_DQ = (swr_req & VGA_CLK) ? {1'b0,mRed[9:5],mGreen[9:5],mBlue[9:5]} : 16'hzzzz;
assign SRAM_WE_N = (swr_req & VGA_CLK) ? 1'b0 : 1'b1;
assign SRAM_ADDR = (swr_req & VGA_CLK) ? {1'b0, VGA_Y[9:2], 1'b0, VGA_X[9:2] - 8'd2} : {sram_raddr_y[8:0], sram_raddr_x[8:0]};

wire [9:0] srVGA_R, srVGA_G, srVGA_B;

wire [9:0] inVGA_R, inVGA_G, inVGA_B;
wire [9:0] inPip_R, inPip_G, inPip_B;
wire [9:0] outPip_R, outPip_G, outPip_B;

wire is_pip;
parameter pip_x_start = 10'd243,
          pip_y_start = 9'd180;
assign sram_raddr_x = VGA_X - pip_x_start;
assign sram_raddr_y = VGA_Y - pip_y_start;
assign is_pip = sram_raddr_x >= 0 && sram_raddr_x <= x_max && sram_raddr_y >= 0 && sram_raddr_y <= y_max;
//assign srVGA_R = (is_pip) ? {SRAM_DQ[14:10], 5'd0} : 10'h3FF;
//assign srVGA_G = (is_pip) ? {SRAM_DQ[9:5], 5'd0} : 10'h0;
//assign srVGA_B = (is_pip) ? {SRAM_DQ[4:0], 5'd0} : 10'h3FF;

assign srVGA_R = (is_pip) ? inPip_R : outPip_R;
assign srVGA_G = (is_pip) ? inPip_G : outPip_G;
assign srVGA_B = (is_pip) ? inPip_B : outPip_B;

assign inVGA_R = (R_in) ? 10'h3FF : 10'h0;
assign inVGA_G = (G_in) ? 10'h3FF : 10'h0;
assign inVGA_B = (B_in) ? 10'h3FF : 10'h0;

wire from_master;
assign from_master = R_in | G_in | B_in;
assign outPip_R = (from_master) ? inVGA_R : Background_Red;
assign outPip_G = (from_master) ? inVGA_G : Background_Green;
assign outPip_B = (from_master) ? inVGA_B : Background_Blue;

assign inPip_R = (from_master) ? inVGA_R : {SRAM_DQ[14:10], 5'd0};
assign inPip_G = (from_master) ? inVGA_G : {SRAM_DQ[9:5], 5'd0};
assign inPip_B = (from_master) ? inVGA_B : {SRAM_DQ[4:0], 5'd0};


/*******************************************
 * Detection Code - Find my bright balls!  *
 *******************************************/

// Top bits of VGA scan point (block location)
wire [6:0] VGA_X_UPPER;
wire [6:0] VGA_Y_UPPER;

// Bottom 3 bits of VGA scan point for 8x8 block size
wire [2:0] VGA_X_LOWER;
wire [2:0] VGA_Y_LOWER;

assign VGA_X_UPPER = VGA_X[9:3];
assign VGA_Y_UPPER = VGA_Y[9:3];
assign VGA_X_LOWER = VGA_X[2:0];
assign VGA_Y_LOWER = VGA_Y[2:0];

// Draw detect1ion box
reg [9:0] red_detbox1, blue_detbox1, green_detbox1;

always @ (*) begin
	if ((VGA_Y_LOWER == 3'd0) && (VGA_Y_UPPER == detbox1_y_top) && (VGA_X_UPPER >= detbox1_x_left) && (VGA_X_UPPER < detbox1_x_right)) begin
		red_detbox1 <= 10'd1023;
		green_detbox1 <= 10'd0;
		blue_detbox1 <= 10'd0;
	end
	else if ((VGA_Y_LOWER == 3'd0) && (VGA_Y_UPPER == detbox1_y_bottom) && (VGA_X_UPPER >= detbox1_x_left) && (VGA_X_UPPER < detbox1_x_right)) begin
		red_detbox1 <= 10'd1023;
		green_detbox1 <= 10'd0;
		blue_detbox1 <= 10'd0;
	end
	else if ((VGA_X_LOWER == 3'd0) && (VGA_X_UPPER == detbox1_x_left) && (VGA_Y_UPPER >= detbox1_y_top) && (VGA_Y_UPPER < detbox1_y_bottom)) begin
		red_detbox1 <= 10'd1023;
		green_detbox1 <= 10'd0;
		blue_detbox1 <= 10'd0;
	end
	else if ((VGA_X_LOWER == 3'd0) && (VGA_X_UPPER == detbox1_x_right) && (VGA_Y_UPPER >= detbox1_y_top) && (VGA_Y_UPPER < detbox1_y_bottom)) begin
		red_detbox1 <= 10'd1023;
		green_detbox1 <= 10'd0;
		blue_detbox1 <= 10'd0;
	end
	else begin
		red_detbox1 <= mRed;
		green_detbox1 <= mGreen;
		blue_detbox1 <= mBlue;
	end
end


reg [59:0] intensity1 [79:0]; // Binary intensity1 matrix
reg scanline1_block; // scanline1 block (block intensity1 is set high if correct intensity1 found in any location within scanline1 block)

// Reset input
wire reset;
assign reset = ~KEY[0];

// Update intensity1 matrix
always @ (posedge VGA_CLK) begin
	if (reset) begin
		scanline1_block <= 1'd0;
	end
	else begin
		if (VGA_X[2:0] != 3'd0) begin
			if ((mRed < 10'd256) && (mBlue < 10'd512) && (mGreen > 10'd512)) begin
				scanline1_block <= 1'd1;
			end
		end
		else begin
			intensity1[VGA_X[9:3]][VGA_Y[9:3]] <= scanline1_block;
			scanline1_block <= 1'd0;
		end
	end
end

// Current scan block locations
reg [6:0] scan1_x, scan1_y;

// intensity1 at previous scan point
reg prev_intensity1;

// scanline1 variables
reg [6:0] scanline1_max_length;
reg [6:0] scanline1_temp_length;
reg [6:0] scanline1_max_length_xend;
reg [6:0] scanline1_max_length_y;

// Variables used temporarily within a scan frame1
reg [6:0] frame1_detect1_x;
reg [6:0] frame1_detect1_xend;
reg [6:0] frame1_detect1_y;
reg [6:0] frame1_detect1_size;

// Output detect1ion variables
reg [6:0] detect1_x;
reg [6:0] detect1_x_mirrored;
reg [6:0] detect1_xend;
reg [6:0] detect1_y;
reg [6:0] detect1_size;

// Used for checking intensity1 slice length at previous detect1ed y location
reg [6:0] prev_detect1_max_length;
reg [6:0] prev_detect1_temp_length;
reg [6:0] prev_detect1_max_length_x;
reg prev_detect1_prev_intensity1;

// Used for drawing detect1ion box
reg [6:0] detbox1_x_left;
reg [6:0] detbox1_y_top;
reg [6:0] detbox1_x_right;
reg [6:0] detbox1_y_bottom;

// detect1 largest blob of light by finding largest equator (contiguous horizontal slice of correct intensity1)
always @ (posedge VGA_CLK) begin
	if (reset) begin // Reset all variables
		scan1_x <= 7'd0;
		scan1_y <= 7'd0;
		prev_intensity1 <= 1'd0;
		scanline1_max_length <= 7'd0;
		scanline1_temp_length <= 7'd0;
		scanline1_max_length_xend <= 7'd0;
		scanline1_max_length_y <= 7'd0;
		frame1_detect1_x <= 7'd0;
		frame1_detect1_xend <= 7'd0;
		frame1_detect1_y <= 7'd0;
		frame1_detect1_size <= 7'd0;
		prev_detect1_max_length <= 7'd0;
		prev_detect1_temp_length <= 7'd0;
		prev_detect1_max_length_x <= 7'd0;
		prev_detect1_prev_intensity1 <= 1'd0;
	end
	else begin
		// Operations within a scanline1
		if (intensity1[scan1_x][scan1_y]) begin // Correct intensity1 detect1ed, increment temporary slice length
			scanline1_temp_length <= scanline1_temp_length + 7'd1;
		end
		else begin // Not the correct intensity1, save this slice as a candidate and restart slice length
			scanline1_temp_length <= 7'd0;
			scanline1_max_length <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length : scanline1_temp_length; // Get length of slice
			scanline1_max_length_y <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length_y : scan1_y; // Get y position of slice
			if (prev_intensity1 == 1'd1) begin // Get x position for end of slice
				scanline1_max_length_xend <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length_xend : scan1_x;
			end
		end
		// Operations for next line and next frame1
		if ((scan1_x == 7'd79) && (scan1_y == 7'd59)) begin // Scanned to end of frame1
		   // Reset variables for next frame1
			scan1_x <= 7'd0;
			scan1_y <= 7'd0;
			scanline1_max_length <= 7'd0;
			scanline1_temp_length <= 7'd0;
			scanline1_max_length_xend <= 7'd0;
			scanline1_max_length_y <= 7'd0;
			// Get y position and size of largest slice in frame1
			frame1_detect1_y <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length_y : scan1_y;
			frame1_detect1_size <= ((scanline1_max_length > scanline1_temp_length) ? scanline1_max_length : scanline1_temp_length);
			// Get x positions for center of largest slice, end of largest slice
			if (intensity1[scan1_x][scan1_y]) begin
				frame1_detect1_x <= (scanline1_max_length > scanline1_temp_length) ? (scanline1_max_length_xend - (scanline1_max_length >> 1) - 7'd1) : (scan1_x - (scanline1_temp_length >> 1) - 7'd1);
				frame1_detect1_xend <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length_xend : scan1_x;
			end
			else begin
				frame1_detect1_x <= scanline1_max_length_xend - (((scanline1_max_length > scanline1_temp_length) ? scanline1_max_length : scanline1_temp_length) >> 1) - 7'd1;
				frame1_detect1_xend <= scanline1_max_length_xend;
			end
			// Update detect1ion if within bounds, use hysteresis: only update for significant difference in intensity1 slice length at previous detect1ed y location or difference significant in center of slice
			if ((frame1_detect1_x >= 7'd1) && (frame1_detect1_x <= 7'd79) && (frame1_detect1_y >= 7'd1) && (frame1_detect1_y <= 7'd59) && ((detect1_size > prev_detect1_max_length + 7'd1) || (prev_detect1_max_length == 7'd0) || (detect1_xend > prev_detect1_max_length_x + 7'd1) || (prev_detect1_max_length_x > detect1_xend + 7'd1))) begin
				detect1_x <= frame1_detect1_x;
				detect1_x_mirrored <= 7'd79 - frame1_detect1_x;
				detect1_xend <= frame1_detect1_xend;
				detect1_y <= frame1_detect1_y;
				detect1_size <= frame1_detect1_size;
				detbox1_x_left <= (frame1_detect1_x > (detect1_size >> 1)) ? frame1_detect1_x - (detect1_size >> 1) : 7'd0;
				detbox1_x_right <= ((frame1_detect1_x + (detect1_size >> 1)) < 7'd79) ? frame1_detect1_x + (detect1_size >> 1) : 7'd79;
				detbox1_y_top <= (frame1_detect1_y > (detect1_size >> 1)) ? frame1_detect1_y - (detect1_size >> 1) : 7'd0;
				detbox1_y_bottom <= ((frame1_detect1_y + (detect1_size >> 1)) < 7'd59) ? frame1_detect1_y + (detect1_size >> 1) : 7'd79;
			end
		end
		else if (scan1_x == 7'd79) begin // Scanned to end of a line
		   // Reset variables for next line
			scan1_x <= 7'd0;
			scan1_y <= scan1_y + 7'd1;
			scanline1_temp_length <= 7'd0;
			// Get y position and size of largest slice in line
			scanline1_max_length <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length : scanline1_temp_length;
			scanline1_max_length_y <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length_y : scan1_y;
			// Get x position for end of largest slice
			if (intensity1[scan1_x][scan1_y]) begin
				scanline1_max_length_xend <= (scanline1_max_length > scanline1_temp_length) ? scanline1_max_length_xend : scan1_x;
			end
		end
		else begin // Otherwise just move the scan to the next block on the right
			scan1_x <= scan1_x + 7'd1;
		end
		// Check intensity1 slice length at previous detect1ed y location
		if ((scan1_y == detect1_y) && (detect1_y != 7'd0)) begin
			if (intensity1[scan1_x][scan1_y]) begin
				prev_detect1_temp_length <= prev_detect1_temp_length + 7'd1;
				if (scan1_x == 7'd0) begin
					prev_detect1_max_length <= 7'd0;
				end
			end
			else begin
				prev_detect1_temp_length <= 7'd0;
				if (scan1_x == 7'd0) begin
					prev_detect1_max_length <= 7'd0;
				end
				else begin
					prev_detect1_max_length <= (prev_detect1_max_length > prev_detect1_temp_length) ? prev_detect1_max_length : prev_detect1_temp_length;
					if (prev_detect1_prev_intensity1 == 1'd1) begin
						prev_detect1_max_length_x <= (prev_detect1_max_length > prev_detect1_temp_length) ? prev_detect1_max_length_x : scan1_x;
					end
				end
			end
			if (scan1_x == 7'd79) begin
				prev_detect1_max_length <= (prev_detect1_max_length > prev_detect1_temp_length) ? prev_detect1_max_length : prev_detect1_temp_length;
				prev_detect1_temp_length <= 7'd0;
				if (intensity1[scan1_x][scan1_y]) begin
					prev_detect1_max_length_x <= (prev_detect1_max_length > prev_detect1_temp_length) ? prev_detect1_max_length_x : scan1_x;
				end
			end
			prev_detect1_prev_intensity1 <= intensity1[scan1_x][scan1_y];
		end
		prev_intensity1 <= intensity1[scan1_x][scan1_y]; 
	end // if not reset
end // detection

endmodule